!BlockClosure class methodsFor: 'documentation' stamp: 'ajh 3/24/2003 15:08'!
aboutClosures"

Package:	Closure Compiler
Author:		Anthony Hannan <ajh18@cornell.edu>
Version:		23 May 2003
Homepage:	http://minnow.cc.gatech.edu/squeak/ClosureCompiler

Prerequisites:
	1.	Squeak 3.4 (http://squeak.org)
	2.	ContextCleanupPlus-ajh
		(http://minnow.cc.gatech.edu/squeak/ClosureCompiler)
	3.	Refactory AST, abstract visitor, and formatter
		(http://www.refactory.com/RefactoringBrowser)
	4.	SmaCC parser generator
		(http://www.refactory.com/Software/SmaCC)

Installing:
	1.	Install ContextCleanupPlus-ajh.
	2.	Install SmaCC Development from SqueakMap Package Loader.
		(the Refactory Browser is included with it).
	3.	File-in the enclosed changesets in their numbered order.
	4.	File-in this ReadMe file to include this description in the image.

Description:

This package adds a closure compiler to the standard Squeak image without replacing the existing compiler. A new 'general' preference called #compileBlocksAsClosures determines which compiler to use for NEW methods only. When true the closure compiler is used, when false the existing compiler is used (default). Block contexts will still be created by existing methods.

The closure compiler is a total rewrite of the Smalltalk compiler using the SmaCC parser generator and the Refactory abstract systax tree. A special thanks goes to John Brant and Don Roberts for creating both the Refactory Browser and Smacc, and to Daniel Vainsencher and Markus Gaelli for porting them, respectively, to Squeak.

For more details on the compiler organization itself see the Compiler class comment. Below is a description of how the compiler creates block closures.

Block Closures

The closure compiler compiles non-inlined blocks as separate CompiledMethods. These block methods are held in the literals of the home method and sent the #createBlock... message at runtime to create BlockClosures. Outer temps referenced by the block are given as args to the #createBlock... message which adds them to the closure it creates. The block closure consists of the method in its sole named field plus its captured temps in its indexable fields. When #value... is sent to a block closure its method is executed in a new MethodContext with the block closure remaining as the receiver. The block method accesses captured temps in indexed fields as if they were named instance variables.

Captured temps that may change after being captured are wrapped in SharedTempHolders and all methods (home and block) access those values indirectly through the holders. Captured temps that may not change (most common) are placed directly in the closure, avoiding the creation and indirection of a holder.

BlockClosures that return to its home context hold the home context as one of its captured temps. Actually, to support safe continuations, it holds the tag of its home context. A ContextTag is a unique object for each context that stays the same for all copies of each context. When a BlockClosure attempts to return to its home context it searches the sender chain for the context that has the right ContextTag and returns to it. This works with copied contexts (continuations) as well as original contexts.

Thanks to captured temps and context tags, block closures are totally separate from their home context. They are reentrant and each activation has its own block-local temps.

BlockClosures are slower primarily because each activation creates a new context, while BlockContexts reuse the same context. An optimizing VM (Jitter, VI4) that stack allocates contexts should rectify this problem. For an example of its slowness (and general slowness of context activation) execute '[(1 to: 1000000) do: [:i | i]] timeToRun' under each compiler (toggle the preference).  On my Pentium II its about 50% slower.

The above example will be unbearably slow under the closure compiler until you rebuild your VM to include a couple of new primitives (186 and 187) added with the closure compiler. They implement BlockClosure>>value... directly. Without them the failure code has to add the closure method to the BlockClosure method dictionary then call #perform:.  A pre-build Windows 3.2.3 VM is on the Closure Compiler home page.


Context Cleanup Plus Changes:

Context Control Enhancements - Methods like #return:, #restart, and #jump have been added to ContextPart for simpler control of the current thread.

Process Control Enhancements - Methods like #completeStep: have been added to Process for simpler control of suspended processes. 'completeStep: aContext' will resume the thread them jump back to the triggering thread when aContext get control back. This allows quick stepping while in the context of the suspended process (i.e. its exception handlers will be active instead of the triggering thread's).

Debugger Enhancements - The debugger uses the new process methods described above.

ContextTags - 'methodContext contextTag' will create a unique ContextTag object for that methodContext and hold it in its receiverMap instance variable which was previously unused. Copies of the methodContext will hold the same contextTag. Holding a contextTag instead of a context directly is safer in light of context copies (see continuations below). A context for a context tag is found in the current thread just be traversing the senders until one is found with that contextTag.

Exceptions Refactored - Exceptions now hold initialContext and handlerContext tags instead of the actual contexts so contexts can be copied safely (described above). Also, searching for handler contexts has been simplified utilizing Context control behavior described above.

Continuations - 'Continuation current' copies thisContext stack and wraps it in a Continuation object. Also, '[:cc | ...] callCC' copies the sender stack into a Continuation and returning it. Evaluating a continuation copies the stack again and resumes that copy. The current process stack is abandoned. When a BlockContext on the stack is copied it points to copied home context if it is also on the stack.

Miscellaneous - Changeset 1 contains general additions to basic objects, collections, and streams, that are used by the rest of changes described above.

Test Cases

Thanks to Boris Gaertner and Rob Withers for writing closure test cases (BlockClosuresTestCase). And thanks to Avi Bryant for writing continuation test cases (ContinuationTest and AmbTest).

"! !
